from .functions import Functions, disabledFunction
from owega.config import debug_print
import json

LTS = Functions()


fcts = {}
fcts["addfun"] = disabledFunction
fcts["delfun"] = disabledFunction
fcts["editfun"] = disabledFunction


def setAdd(addfun):
	fcts["addfun"] = addfun


def setDel(delfun):
	fcts["delfun"] = delfun


def setEdit(editfun):
	fcts["editfun"] = editfun


# add_memory(new_memory: str) -> index
def __add_memory(*args, **kwargs):
	new_memory = ""
	if len(args) > 0:
		new_memory = args[0]
	new_memory = kwargs.get("new_memory", new_memory)
	rdict = {}
	if not new_memory:
		rdict["function_status"] = "No memory provided, memory not added"
		return json.dumps(rdict)
	rdict["new_memory_index"] = fcts["addfun"](new_memory)
	if rdict["new_memory_index"] < 0:
		rdict = {}
		rdict["function_status"] = "Something went wrong, no memory added"
		return json.dumps(rdict)
	rdict["function_status"] = ("New memory added at index "
		+ f"{rdict['new_memory_index']}")
	return json.dumps(rdict)


__add_memory_desc = {
	"name": "add_memory",
	"description": "Adds a new souvenir to the AI's long-term memory",
	"parameters": {
		"type": "object",
		"properties": {
			"new_memory": {
				"type": "string",
				"description": "the memory to remember, as a sentence from "
				+ "the AI's point of view, not from the user's"
			}
		},
		"required": ["new_memory"],
	},
}


LTS.addFunction(__add_memory, __add_memory_desc)


# remove_memory(index_to_delete: int) -> content
def __remove_memory(*args, **kwargs):
	index_to_delete = -1
	if len(args) > 0:
		index_to_delete = args[0]
	index_to_delete = kwargs.get("index_to_delete", index_to_delete)
	rdict = {}
	if index_to_delete < 0:
		rdict["function_status"] = \
			"No memory index provided, memory not removed"
		return json.dumps(rdict)
	rdict["memory_deleted"] = fcts["delfun"](index_to_delete)
	rdict["function_status"] = "Old memory deleted!"
	return json.dumps(rdict)


__remove_memory_desc = {
	"name": "remove_memory",
	"description": "Removes a souvenir from the AI's long-term memory",
	"parameters": {
		"type": "object",
		"properties": {
			"index_to_delete": {
				"type": "integer",
				"description": "the index to the memory to forget"
			}
		},
		"required": ["index_to_delete"],
	},
}


LTS.addFunction(__remove_memory, __remove_memory_desc)


# edit_memory(index_to_edit: int, new_memory: str) -> false/true
def __edit_memory(*args, **kwargs):
	index_to_edit = -1
	new_memory = ""
	if len(args) > 0:
		index_to_edit = args[0]
		if len(args) > 1:
			new_memory = args[1]
	index_to_edit = kwargs.get("index_to_edit", index_to_edit)
	new_memory = kwargs.get("new_memory", new_memory)
	rdict = {}
	if (not new_memory) or (index_to_edit < 0):
		rdict["function_status"] = "Missing argument, no memory were edited"
		return json.dumps(rdict)
	rval = fcts["editfun"](index_to_edit, new_memory)
	if not rval:
		rdict = {}
		rdict["function_status"] = \
			"Something went wrong, no memory were edited"
		return json.dumps(rdict)
	rdict["function_status"] = "Memory edited successfully!"
	return json.dumps(rdict)


__edit_memory_desc = {
	"name": "edit_memory",
	"description": "Edits an existing souvenir from the AI's long-term memory",
	"parameters": {
		"type": "object",
		"properties": {
			"index_to_edit": {
				"type": "integer",
				"description": "the index to the memory to edit"
			},
			"new_memory": {
				"type": "string",
				"description": "the new memory to be remembered instead of the"
				+ "old one, as a sentence from the AI's point of view, "
				+ "not the user's"
			},
		},
		"required": ["index_to_edit", "new_memory"],
	},
}


LTS.addFunction(__edit_memory, __edit_memory_desc)



import editor
import prompt_toolkit as pt
from ..utils import (
	info_print,
	clrtxt,
)
from ..OwegaSession import OwegaSession as ps


def handle_edit(temp_file, messages, given="", temp_is_temp=False):
	for msg_id, msg in enumerate(messages.messages):
		role = msg.get('role', 'unknown')
		if role == 'user':
			print(f"[\033[0;93mUSER\033[0m] [\033[0;92m{msg_id}\033[0m]:")
			print('\033[0;37m', end='')
			print(
				msg.get('content', '')
				.encode('utf16', 'surrogatepass')
				.decode('utf16')
			)
			print('\033[0m', end='')
			print()
		elif role == 'system':
			print(f"[\033[0;95mSYSTEM\033[0m] [\033[0;92m{msg_id}\033[0m]:")
			print('\033[0;37m', end='')
			print(
				msg.get('content', '')
				.encode('utf16', 'surrogatepass')
				.decode('utf16')
			)
			print('\033[0m', end='')
			print()
		elif role == 'assistant':
			print(f"[\033[0;96mOWEGA\033[0m] [\033[0;92m{msg_id}\033[0m]:")
			print('\033[0;37m', end='')
			print(
				msg.get('content', '')
				.encode('utf16', 'surrogatepass')
				.decode('utf16')
			)
			print('\033[0m', end='')
			print()
	try:
		msg_id = ps['integer'].prompt(pt.ANSI(
			'\n' + clrtxt("magenta", " message ID ") + ': ')).strip()
	except (ValueError, KeyboardInterrupt, EOFError):
		info_print("Invalid message ID, cancelling edit")
		return messages

	msg_id = int(msg_id)

	if (msg_id < 0) or (msg_id >= len(messages.messages)):
		info_print("Invalid message ID, cancelling edit")
		return messages

	try:
		new_msg = editor.edit(
			contents=messages.messages[msg_id].get('content', '').encode(
				'utf16', 'surrogatepass')
		).decode('utf16')
	except UnicodeDecodeError:
		info_print("Error handling given message, edit not saved")
		return messages
	if new_msg:
		messages.messages[msg_id]['content'] = new_msg
	else:
		info_print("Message empty, deleting it...")
		messages.messages.pop(msg_id)

	return messages


item_edit = {
	"fun": handle_edit,
	"help": "edits the history",
	"commands": ["edit"],
}



import os
import time
import tempfile
import openai
import prompt_toolkit as pt
from ..config import baseConf
from ..utils import (
	clrtxt,
	estimated_tokens,
	info_print,
	play_opus,
)
from ..ask import ask
from ..OwegaFun import existingFunctions, functionlist_to_toollist
from ..OwegaSession import OwegaSession as ps


def guess_language(file_path: str) -> str:
	filename = file_path.split('/')[-1].lower()
	if 'iptables' in filename:
		return 'iptables'
	if 'makefile' in filename:
		return 'makefile'

	ext = filename.split('.')[-1]

	known_exts = {}

	def add_exts(lang: str, exts) -> None:
		if isinstance(exts, str):
			known_exts[exts] = lang.lower()
		else:
			for ext in exts:
				known_exts[ext] = lang.lower()

	def add_sexts(langs) -> None:
		if isinstance(langs, str):
			known_exts[langs] = langs.lower()
		else:
			for lang in langs:
				known_exts[lang] = lang.lower()

	add_exts('brainfuck', ['brainfuck', 'bf'])
	add_exts('c', ['c', 'h'])
	add_exts(
		'cpp',
		['cpp', 'cxx', 'c++', 'cc', 'hpp', 'hxx', 'h++', 'hh']
	)
	add_exts('csharp', 'cs')
	add_exts('clojure', ['clojure', 'clj'])
	add_exts('diff', ['diff', 'patch'])
	add_exts('dos', ['dos', 'bat', 'cmd'])
	add_exts('elixir', ['elixir', 'ex', 'exs'])
	add_exts('erlang', ['erlang', 'erl'])
	add_exts('fortran', ['fortran', 'f90', 'f95'])
	add_exts('golang', 'go')
	add_exts('xml',
		['xml', 'rss', 'atom', 'xjb', 'xsd', 'xsl', 'plist', 'svg'])
	add_exts('haskell', ['haskell', 'hs'])
	add_exts('haxe', ['haxe', 'hx'])
	add_exts('julia', ['julia', 'jl'])
	add_exts('kotlin', ['kotlin', 'kt'])
	add_exts('makefile', ['mk', 'mak', 'make', 'makefile'])
	add_exts('markdown', ['markdown', 'md', 'mkdown', 'mkd'])
	add_exts('mirc', ['mirc', 'mrc'])
	add_exts('nim', ['nim', 'nimrod'])
	add_exts('ocaml', ['ocaml', 'ml'])
	add_exts('objc',
		['objectivec', 'mm', 'objc', 'obj-c', 'obj-c++', 'objective-c++'])
	add_exts('openscad', ['openscad', 'scad'])
	add_exts('perl', ['perl', 'pl', 'pm'])
	add_exts('pgsql', ['pgsql', 'postgres', 'postgresql'])
	add_exts('powershell', ['powershell', 'ps', 'ps1'])
	add_exts('puppet', ['puppet', 'pp'])
	add_exts('python', ['python', 'py', 'gyp'])
	add_exts('rpm', ['rpm-specfile', 'rpm', 'spec', 'rpm-spec', 'specfile'])
	add_exts('ruby', ['ruby', 'rb', 'gemspec', 'podspec', 'thor', 'irb'])
	add_exts('rust', ['rust', 'rs'])
	add_exts('scilab', ['scilab', 'sci'])
	add_exts('tcl', ['tcl', 'tk'])
	add_exts('typescript', ['typescript', 'ts', 'tsx', 'mts', 'cts'])
	add_exts('verilog', ['verilog', 'v'])
	add_exts('yaml', ['yaml', 'yml'])
	add_sexts([
		'ada', 'ino', 'awk', 'sh', 'bash', 'zsh', 'basic', 'bbcode', 'blade',
		'bnf', 'bqn', 'cmake', 'cobol', 'coq', 'csp', 'css', 'd', 'dart',
		'fix', 'gcode', 'gdscript', 'gradle', 'graphql', 'gsql', 'html',
		'xhtml', 'haml', 'hlsl', 'ini', 'toml', 'json', 'java', 'js', 'jsx',
		'tex', 'leaf', 'less', 'lisp', 'lua', 'matlab', 'nix', 'oak', 'ocl',
		'glsl', 'pf', 'php', 'parser3', 'processing', 'prolog', 'properties',
		'qml', 'r', 'sas', 'scss', 'sql', 'scala', 'scheme', 'sfz', 'smali',
		'spl', 'svelte', 'swift', 'thrift', 'toit', 'tp', 'tsql', 'vbnet',
		'vb', 'vba', 'vbscript', 'vbs', 'vhdl', 'vala'
	])

	lang = known_exts.get(ext.lower(), '')

	if not lang:
		if 'vim' in filename:
			return 'vim'

	return lang


# get input from a file instead of the terminal
def handle_finput(temp_file, messages, given="", temp_is_temp=False):
	given = given.strip()
	given = given.split(' ')[0]
	if given:
		file_path = given
	else:
		file_path = ps['load'].prompt(pt.ANSI(
			clrtxt("yellow", " FILE LOCATION ") + ": ")).strip()
	if (os.path.exists(file_path)):
		user_prompt = ps['main'].prompt(pt.ANSI(
			clrtxt("yellow", " PRE-FILE PROMPT ") + ": ")).strip()
		with open(file_path, "r") as f:
			language = guess_language(file_path)
			file_contents = f.read()
			full_prompt = \
				f"{user_prompt}\n```{language}\n{file_contents}\n```\n"
			if baseConf.get("estimation", False):
				etkn = estimated_tokens(
					full_prompt,
					messages,
					functionlist_to_toollist(existingFunctions.getEnabled())
				)
				cost_per_token = (
					0.03
					if 'gpt-4' in baseConf.get("model", "")
					else 0.003
				) / 1000
				cost = cost_per_token * etkn
				print(f"\033[37mestimated tokens: {etkn}\033[0m")
				print(f"\033[37mestimated cost: {cost:.5f}\033[0m")
			if baseConf.get("debug", False):
				pre_time = time.time()
			messages = ask(
				prompt=full_prompt,
				messages=messages,
				model=baseConf.get("model", ""),
				temperature=baseConf.get("temperature", 0.8),
				max_tokens=baseConf.get("max_tokens", 3000),
				top_p=baseConf.get("top_p", 1.0),
				frequency_penalty=baseConf.get("frequency_penalty", 0.0),
				presence_penalty=baseConf.get("presence_penalty", 0.0)
			)
			if baseConf.get("debug", False):
				post_time = time.time()
				print(f"\033[37mrequest took {post_time-pre_time:.3f}s\033[0m")
			print()
			print(' ' + clrtxt("magenta", " Owega ") + ": ")
			print()
			print(messages.last_answer())
			if baseConf.get('tts_enabled', False):
				tmpfile = tempfile.NamedTemporaryFile(
					prefix="owegatts.",
					suffix=".opus",
					delete=False
				)
				tmpfile.close()
				tts_answer = openai.audio.speech.create(
					model='tts-1',
					voice='nova',
					input=messages.last_answer()
				)
				if baseConf.get("debug", False):
					posttts_time = time.time()
					print(f"\033[37mrequest took {posttts_time-pre_time:.3f}s\033[0m")
				tts_answer.stream_to_file(tmpfile.name)
				play_opus(tmpfile.name)
				os.remove(tmpfile.name)
	else:
		info_print(f"Can't access {file_path}")
	return messages


item_finput = {
	"fun": handle_finput,
	"help": "sends a prompt and a file from the system",
	"commands": ["file_input"],
}



import prompt_toolkit as pt
from ..utils import (
	info_print,
	clrtxt,
)
from ..OwegaSession import OwegaSession as ps


# adds a system message
def handle_del_sysmem(temp_file, messages, given="", temp_is_temp=False):
	for index, sysmem in enumerate(messages.systemsouv):
		print(f"[\033[0;95mSystem souvenir\033[0m] [\033[0;92m{index}\033[0m]:")
		print('\033[0;37m', end='')
		print(sysmem)
		print('\033[0m', end='')
		print()
	try:
		msg_id = ps['integer'].prompt(pt.ANSI(
			'\n' + clrtxt("magenta", " message ID ") + ': ')).strip()
	except (ValueError, KeyboardInterrupt, EOFError):
		info_print("Invalid message ID, cancelling edit")
		return messages

	msg_id = int(msg_id)

	if (msg_id < 0) or (msg_id >= len(messages.systemsouv)):
		info_print("Invalid message ID, cancelling edit")
		return messages

	messages.systemsouv.pop(msg_id)

	return messages


item_del_sysmem = {
	"fun": handle_del_sysmem,
	"help": "deletes a system souvenir",
	"commands": ["del_sysmem"],
}



import prompt_toolkit as pt
from ..config import baseConf
from ..utils import (info_print, clrtxt)
from ..OwegaSession import OwegaSession as ps


# change requested tokens amount
def handle_tokens(temp_file, messages, given="", temp_is_temp=False):
	given = given.strip()
	if given.isdigit():
		baseConf["max_tokens"] = int(given)
		info_print(f'Set requested tokens to {baseConf.get("max_tokens", 3000)}')
		return messages
	info_print(f'Currently requested tokens: {baseConf.get("max_tokens", 3000)}')
	info_print('How many tokens should be requested?')
	new_tokens = pt.prompt(pt.ANSI(
		'\n' + clrtxt("magenta", " tokens ") + ': '
	)).strip()
	if new_tokens.isdigit():
		baseConf["max_tokens"] = int(new_tokens)
		info_print(f'Set requested tokens to {baseConf.get("max_tokens", 3000)}')
	else:
		info_print('Invalid input, keeping current requested tokens amount')
	return messages


item_tokens = {
	"fun": handle_tokens,
	"help": "changes the amount of requested tokens",
	"commands": ["tokens"],
}



import prompt_toolkit as pt
from ..config import baseConf
from ..utils import (
	info_print,
	clrtxt,
)
from ..OwegaSession import OwegaSession as ps


# change frequency penalty
def handle_frequency(temp_file, messages, given="", temp_is_temp=False):
	given = given.strip()
	try:
		new_frequency_penalty = float(given)
	except ValueError:
		info_print('Current frequency penalty: '
			+ f'{baseConf.get("frequency_penalty", 1.0)}')
		info_print('New frequency penalty value (0.0 - 2.0, defaults 0.0)')
		try:
			new_frequency_penalty = ps['float'].prompt(pt.ANSI(
				'\n' + clrtxt("magenta", " frequency penalty ") + ': ')).strip()
		except (ValueError, KeyboardInterrupt, EOFError):
			info_print("Invalid frequency penalty.")
			return messages
	baseConf["frequency_penalty"] = float(new_frequency_penalty)
	nv = baseConf.get('frequency_penalty', 0.0)
	if nv > 2.0:
		info_print('Penalty too high, capping to 2.0')
		baseConf["frequency_penalty"] = 2.0
	if nv < 0.0:
		info_print('Penalty too low, capping to 0.0')
		baseConf["frequency_penalty"] = 0.0
	info_print('Set frequency penalty to '
		+ f'{baseConf.get("frequency_penalty", 0.0)}')
	return messages


item_frequency = {
	"fun": handle_frequency,
	"help": "sets the frequency penalty (0.0 - 2.0, defaults 0.0)",
	"commands": ["frequency"],
}



from .changelogEntry import ChangelogEntry
from .version import Version


class Changelog:
	def __init__(self):
		self.logs = []
		self.log = ""
		self.version = Version(0, 0, 0)
		self.initLogs()
		self.genLog()

	def initLogs(self):
		self.logs.append(
			ChangelogEntry(4, 12, 8)
			.addLine("Added a vim modeline to history files")
			.addLine("  to specify it's json5, not json.")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 7)
			.addLine('Fixed a minor bug where /file_input would insert a "\'"')
			.addLine("  after the file contents.")
			.addLine("Also, added filetype information on codeblocks with")
			.addLine("  /file_input, depending on the file extension")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 6)
			.addLine("Fixed emojis crashing the edit function because utf16")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 5)
			.addLine("Fixed emojis crashing the history because utf16")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 4)
			.addLine("Fixed requirements to use json5 instead of json-five")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 3)
			.addLine("Fixed requirements to be more lenient")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 2)
			.addLine("Fixed TUI-mode TTS")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 1)
			.addLine("Added -e/--estimate option to estimate consumption")
		)
		self.logs.append(
			ChangelogEntry(4, 12, 0)
			.addLine("Added -T/--training option to generate training line")
		)

		self.logs.append(
			ChangelogEntry(4, 11, 5)
			.addLine("Fixed requirements in setup.py not working when getting")
			.addLine("the source from PyPI")
		)
		self.logs.append(
			ChangelogEntry(4, 11, 4)
			.addLine("Fixed edit with blank message (remove message)")
		)
		self.logs.append(
			ChangelogEntry(4, 11, 3)
			.addLine("Fixed /genconf")
		)
		self.logs.append(
			ChangelogEntry(4, 11, 2)
			.addLine("Fixed -a / single_ask")
		)
		self.logs.append(
			ChangelogEntry(4, 11, 1)
			.addLine("Oops, last version broke owega, fixed here")
			.addLine("(Problem was I forgot to export submodules in setup.py)")
		)
		self.logs.append(
			ChangelogEntry(4, 11, 0)
			.addLine("Huge refactor, added TTS as config parameter")
		)

		self.logs.append(
			ChangelogEntry(4, 10, 3)
			.addLine("- changed from OpenAI to Owega in term display")
		)
		self.logs.append(
			ChangelogEntry(4, 10, 2)
			.addLine("- added cost estimation in token estimation")
		)
		self.logs.append(
			ChangelogEntry(4, 10, 1)
			.addLine("- added support server in readme and pypi")
		)
		self.logs.append(
			ChangelogEntry(4, 10, 0)
			.addLine("- added system souvenirs (add_sysmem/del_sysmem)")
		)

		self.logs.append(
			ChangelogEntry(4, 9, 0)
			.addLine("- added system command")
		)

		self.logs.append(
			ChangelogEntry(4, 8, 2)
			.addLine("- added infos to pypi page")
			.addLine("- changed to automatic script generation (setup.py)")
		)
		self.logs.append(
			ChangelogEntry(4, 8, 1)
			.addLine("Oops, forgot to add requirements to setup.py")
			.addLine("Automated the process, should be good now")
		)
		self.logs.append(
			ChangelogEntry(4, 8, 0)
			.addLine("Edit update")
			.addLine("- you can now edit the history from the TUI")
			.addLine("- on a side note, I also improved completion for files")
			.addLine("    and numeric values (temperature, top_p, penalties...)")
		)

		self.logs.append(
			ChangelogEntry(4, 7, 3)
			.addLine("Added ctrl+C handling when playing TTS to stop speaking.")
		)
		self.logs.append(
			ChangelogEntry(4, 7, 2)
			.addLine(
				"Fixed a bug where the output tts file could not be set to mp3"
			)
			.addLine("  (it was previously checking for mp4 extension, lol)")
		)
		self.logs.append(
			ChangelogEntry(4, 7, 1)
			.addLine("Now prints message before reading TTS")
			.addLine("Also, removes the pygame init message")
		)
		self.logs.append(
			ChangelogEntry(4, 7, 0)
			.addLine("Added TTS (using pygame)")
		)

		self.logs.append(
			ChangelogEntry(4, 6, 2)
			.addLine("Oops, forgot to check help, help should be fixed now")
		)
		self.logs.append(
			ChangelogEntry(4, 6, 1)
			.addLine("Added support for overwriting config file")
		)
		self.logs.append(
			ChangelogEntry(4, 6, 0)
			.addLine("Fine tweaking update")
			.addLine("- added command for changing the temperature")
			.addLine("- added top_p command and parameter")
			.addLine("- added frequency penalty command and parameter")
			.addLine("- added presence penalty command and parameter")
			.addLine("- fixed /quit and /exit not working")
			.addLine("- fixed tab completion")
		)

		self.logs.append(
			ChangelogEntry(4, 5, 3)
			.addLine("Fixed files being removed everytime")
		)
		self.logs.append(
			ChangelogEntry(4, 5, 2)
			.addLine("Now removes temp files even if ctrl+c if they are empty")
		)
		self.logs.append(
			ChangelogEntry(4, 5, 1)
			.addLine(
				"fixed owega bash script for systems that still have "
				+ "PYTHON 2 AS DEFAULT"
			)
			.addLine("WTF GUYS GET OVER IT, IT'S BEEN DEPRECATED SINCE 2020")
		)
		self.logs.append(
			ChangelogEntry(4, 5, 0)
			.addLine("Added support for organization specification")
		)

		self.logs.append(
			ChangelogEntry(4, 4, 0)
			.addLine("Changed from json to json5 (json-five)")
		)

		self.logs.append(
			ChangelogEntry(4, 3, 6)
			.addLine(
				"Re-added handling of invalid request, mostly for too large "
				+ "requests"
			)
		)
		self.logs.append(
			ChangelogEntry(4, 3, 5)
			.addLine("Added exception handling for token estimation")
		)
		self.logs.append(
			ChangelogEntry(4, 3, 4)
			.addLine("Re-added server unavailable error handling")
		)
		self.logs.append(
			ChangelogEntry(4, 3, 3)
			.addLine("Changed time taken to only show up to ms")
		)
		self.logs.append(
			ChangelogEntry(4, 3, 2)
			.addLine("Fixed 4.3.1 :p")
		)
		self.logs.append(
			ChangelogEntry(4, 3, 1)
			.addLine("Added time taken per request in debug output")
		)
		self.logs.append(
			ChangelogEntry(4, 3, 0)
			.addLine("Added token estimation")
		)

		self.logs.append(
			ChangelogEntry(4, 2, 0)
			.addLine("VERY IMPORTANT UPDATE: NOW COMPATIBLE WITH OPENAI 1.1.1")
		)

		self.logs.append(
			ChangelogEntry(4, 1, 1)
			.addLine("Removed a warning due to beautifulsoup4")
		)
		self.logs.append(
			ChangelogEntry(4, 1, 0)
			.addLine("Changed the getpage function to strip the text")
		)

		self.logs.append(
			ChangelogEntry(4, 0, 4)
			.addLine("Fixed context not working correctly")
		)
		self.logs.append(
			ChangelogEntry(4, 0, 3)
			.addLine("Added README to pypi page")
		)
		self.logs.append(
			ChangelogEntry(4, 0, 2)
			.addLine("Fixed a typo where owega wouldn't send the memory")
		)
		self.logs.append(
			ChangelogEntry(4, 0, 1)
			.addLine(
				"oops, forgot to change the setup.py and now I messed up my "
				+ "4.0.0! >:C"
			)
		)
		self.logs.append(
			ChangelogEntry(4, 0, 0)
			.addLine("LTS: Long-Term-Souvenirs")
			.addLine("The AI now have long-term memory!!!")
			.addLine(
				"Huge update: full refactoring, the code is now readable!")
			.addLine(
				"Also, the name is now Owega (it's written with unicode "
				+ "characters though)"
			)
			.addLine(
				"You can see the new code here: "
				+ "https://git.pyrokinesis.fr/darkgeem/owega"
			)
			.addLine(
				"Also, the project is now available on PyPI so, just go pip "
				+ "install owega!"
			)
		)

		self.logs.append(
			ChangelogEntry(3, 9, 4)
			.addLine("changed default values")
		)
		self.logs.append(
			ChangelogEntry(3, 9, 3)
			.addLine("fixed api key not saving with /genconf")
		)
		self.logs.append(
			ChangelogEntry(3, 9, 2)
			.addLine(
				"changed the temp file creation method for non-unix systems")
		)
		self.logs.append(
			ChangelogEntry(3, 9, 1)
			.addLine(
				"fixed an issue when the openai api key does not exist "
				+ "anywhere"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 9, 0)
			.addLine("Windows update")
			.addLine("  - Do I really need to explain that update?")
		)
		self.logs.append(
			ChangelogEntry(3, 8, 1)
			.addLine("added a debug option for devs")
		)
		self.logs.append(
			ChangelogEntry(3, 8, 0)
			.addLine("WEB download update")
			.addLine(
				"  - added a get_page function for openchat to get pages "
				+ "without the need"
			)
			.addLine("      for curl")
		)
		self.logs.append(
			ChangelogEntry(3, 7, 0)
			.addLine("DIRECT CONTEXT COMMANDS update:")
			.addLine(
				"  - now, you can use commands in one line, instead of waiting"
				+ "for prompt"
			)
			.addLine("      example: /save hello.json")
			.addLine(
				"      (instead of typing /save, then enter, then typing "
				+ "hello.json"
			)
			.addLine(
				"       works on all commands, the only specific case being "
				+ "file_input.)"
			)
			.addLine(
				"  - file_input as a direct command takes only one argument: "
				+ "the file"
			)
			.addLine(
				"      to load (e.g. /load ./src/main.c). The pre-prompt will "
				+ "be asked"
			)
			.addLine(
				"      directly instead of having to do it in three steps")
			.addLine("        (/load, then filename, then pre-prompt)")
			.addLine(
				"  - also, fixed /tokens splitting the prompt instead of the "
				+ "user input"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 6, 0)
			.addLine("PREFIX update:")
			.addLine(
				"  - added prefixes for command (changeable in the config)")
			.addLine(
				"  - reformatted most of the main loop code to split it in "
				+ "handlers"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 5, 2)
			.addLine("added infos on bottom bar")
		)
		self.logs.append(
			ChangelogEntry(3, 5, 1)
			.addLine(
				'added "commands" command, to enable/disable command '
				+ 'execution'
			)
		)
		self.logs.append(
			ChangelogEntry(3, 5, 0)
			.addLine(
				"WEB update: now added a flask app, switched repos to its own")
		)
		self.logs.append(
			ChangelogEntry(3, 4, 0)
			.addLine("CLI update:")
			.addLine(
				"  - added command-line options to change input/output files")
			.addLine(
				"  - added command-line option to ask a question from command "
				+ "line"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 3, 1)
			.addLine(
				"added tokens command, to change the amount of requested "
				+ "tokens"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 3, 0)
			.addLine(
				"implemented prompt_toolkit, for better prompt handling, "
				+ "newlines with"
			)
			.addLine("control+n")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 11)
			.addLine(
				"now, the default gpt models implement function calling, no "
				+ "need for"
			)
			.addLine("0613 anymore")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 10)
			.addLine(
				"added a command line option for specifying the config file")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 9)
			.addLine(
				"changed execute's subprocess call to shell=True, now "
				+ "handling pipes..."
			)
		)
		self.logs.append(
			ChangelogEntry(3, 2, 8)
			.addLine("fixed command execution stderr handling")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 7)
			.addLine(
				"fixed json sometimes not correctly formatted when writing "
				+ "multiple lines"
			)
			.addLine("files")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 6)
			.addLine(
				"reversed the changelog order, fixed function calling chains")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 5)
			.addLine(
				"now handling non-zero exit status when running a command")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 4, "fix1")
			.addLine("fixed a missing parenthesis")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 4)
			.addLine(
				"fixed variables and ~ not expanding when executing a command")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 3)
			.addLine("added create_file as a function OpenAI can call")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 2)
			.addLine(
				"fixed openchat sometimes not detecting the command has been "
				+ "ran"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 2, 1)
			.addLine("fixed a space missing in openchat's function calling")
		)
		self.logs.append(
			ChangelogEntry(3, 2, 0)
			.addLine("FUNCTION CALLING UPDATE:")
			.addLine(
				"added function calling, now openchat is able to run commands")
			.addLine("on your computer, as long as you allow it to")
			.addLine(
				"(you will be prompted on each time it tries to run a "
				+ "command)"
			)
			.addLine(
				"!!! only available on -0613 models (gpt-3.5-turbo-0613, "
				+ "gpt-4-0613) !!!"
			)
			.addLine(
				"will be available on all gpt models from 2023-06-27, with "
				+ "the latest"
			)
			.addLine("openchat 3.2.X patch")
		)
		self.logs.append(
			ChangelogEntry(3, 1, 1)
			.addLine("now handling the service unavailable error")
		)
		self.logs.append(
			ChangelogEntry(3, 1, 0)
			.addLine("BMU (Better Module Update)!")
			.addLine("modified MSGS:")
			.addLine("  - added last_question()")
			.addLine("  - changed last_answer()")
			.addLine("modified ask() to allow for blank prompt,")
			.addLine("  which will reuse the last question")
		)
		self.logs.append(
			ChangelogEntry(3, 0, 3)
			.addLine(
				"quitting with EOF will now discard the temp file (^C will "
				+ "still keep it)"
			)
		)
		self.logs.append(
			ChangelogEntry(3, 0, 2)
			.addLine("added conversion script")
		)
		self.logs.append(
			ChangelogEntry(3, 0, 1)
			.addLine("added changelog")
		)
		self.logs.append(
			ChangelogEntry(3, 0, 0)
			.addLine("changed conversation save from pickle to json")
		)

		self.logs.append(
			ChangelogEntry(2, 2, 4)
			.addLine("automatic temp file save")
		)
		self.logs.append(
			ChangelogEntry(2, 2, 3)
			.addLine(
				"genconf now saves the current conf instead of a blank "
				+ "template"
			)
		)
		self.logs.append(
			ChangelogEntry(2, 2, 2)
			.addLine(
				"stripped user input (remove trailing spaces/tabs/newlines)")
		)
		self.logs.append(
			ChangelogEntry(2, 2, 1)
			.addLine(
				"added license and version info in command line (-l and -v)")
		)
		self.logs.append(
			ChangelogEntry(2, 2, 0)
			.addLine("added context command to change GPT's definition")
		)
		self.logs.append(
			ChangelogEntry(2, 1, 1)
			.addLine("added file_input in help command")
		)
		self.logs.append(
			ChangelogEntry(2, 1, 0)
			.addLine("added file_input command")
		)
		self.logs.append(
			ChangelogEntry(2, 0, 1)
			.addLine("added genconf command")
		)
		self.logs.append(
			ChangelogEntry(2, 0, 0)
			.addLine("WTFPL license")
		)

	def genLog(self):
		self.logs.sort()
		self.version = self.logs[-1].version
		self.version_str = str(self.logs[-1].version)
		self.log = f"OWEGA v{self.version_str} CHANGELOG:"
		for entry in self.logs:
			ver = entry.version
			if (not ver.status) and ver.patch == 0:
				self.log += '\n'
				if ver.minor == 0:
					self.log += '\n'
			self.log += '\n'
			if 'rc' in ver.status:
				self.log += '\033[91m'
			self.log += str(entry)
			if 'rc' in ver.status:
				self.log += '\033[m'


OwegaChangelog = Changelog()



import prompt_toolkit as pt
from ..config import baseConf
from ..changelog import OwegaChangelog


# bottom toolbar and style for prompt_toolkit
def main_bottom_toolbar(what: str = "toolbar"):
	if what == "style":
		msd = {
			'red':     '#000000 bg:#FF0000',  # noqa: E241
			'green':   '#000000 bg:#00FF00',  # noqa: E241
			'blue':    '#000000 bg:#0000FF',  # noqa: E241
			'yellow':  '#000000 bg:#FFFF00',  # noqa: E241
			'magenta': '#000000 bg:#FF00FF',  # noqa: E241
			'cyan':    '#000000 bg:#00FFFF',  # noqa: E241
			'white':   '#000000 bg:#FFFFFF',  # noqa: E241
		}
		msd['bottom-toolbar'] = msd['white']
		msd['bottom-even'] = msd['magenta']
		msd['bottom-odd'] = msd['cyan']
		msd['bottom-on'] = msd['green']
		msd['bottom-off'] = msd['red']
		main_style = pt.styles.Style.from_dict(msd)
		return main_style

	class tr:
		def __init__(self):
			self.table = []
			self.count = 0

		def add(self, prefix="/", txt="", color="yellow"):
			new_class = "white"
			if not isinstance(txt, str):
				txt = str(txt)
			if self.count:
				self.table.append((
					"class:blue",
					" - "
				))
			if prefix:
				if txt:
					prefix = prefix + ": "
				self.table.append((
					f"class:{new_class}",
					prefix
				))
			if txt:
				self.table.append((
					f"class:{color}",
					txt
				))
			self.count += 1

	to_ret = tr()
	to_ret.add(f"v{OwegaChangelog.version}")
	to_ret.add("model", baseConf.get("model", "unknown"))
	to_ret.add(
		"cmds",
		"ON" if baseConf.get("commands") else "OFF",
		"bottom-on" if baseConf.get("commands") else "bottom-off"
	)
	to_ret.add("tokens", baseConf.get("max_tokens", "unknown"))
	to_ret.add(
		"estm",
		"ON" if baseConf.get("estimation") else "OFF",
		"bottom-on" if baseConf.get("estimation") else "bottom-off"
	)
	to_ret.add("temp", baseConf.get("temperature", "unknown"))
	to_ret.add("top_p", baseConf.get("top_p", "unknown"))
	to_ret.add("freq.plty", baseConf.get("frequency_penalty", "unknown"))
	to_ret.add("pr.plty", baseConf.get("presence_penalty", "unknown"))
	to_ret.add(
		"TTS",
		"ON" if baseConf.get("tts_enabled") else "OFF",
		"bottom-on" if baseConf.get("tts_enabled") else "bottom-off"
	)
	return to_ret.table



from ..utils import print_help
from .handle_quit import item_quit
from .handle_genconf import item_genconf
from .handle_history import item_history
from .handle_commands import item_commands
from .handle_estimation import item_estimation
from .handle_tokens import item_tokens
from .handle_context import item_context
from .handle_save import item_save
from .handle_load import item_load
from .handle_model import item_model
from .handle_finput import item_finput
from .handle_temperature import item_temperature
from .handle_top_p import item_top_p
from .handle_frequency import item_frequency
from .handle_presence import item_presence
from .handle_edit import item_edit
from .handle_system import item_system
from .handle_add_sysmem import item_add_sysmem
from .handle_del_sysmem import item_del_sysmem
from .handle_tts import item_tts
from ..OwegaSession import OwegaSession as ps
from ..handlerBase import items, handlers, handler_helps


# prints help
def handle_help(temp_file, messages, given="", temp_is_temp=False):
	print_help(handler_helps)
	return messages


item_help = {
	"fun": handle_help,
	"help": "Shows this help",
	"commands": ["help"],
}


def void_func(temp_file, messages, given="", temp_is_temp=False):
	return messages


def populate():
	items.append(item_help)
	items.append(item_quit)
	items.append(item_genconf)
	items.append(item_history)
	items.append(item_commands)
	items.append(item_estimation)
	items.append(item_tokens)
	items.append(item_context)
	items.append(item_save)
	items.append(item_load)
	items.append(item_model)
	items.append(item_finput)
	items.append(item_temperature)
	items.append(item_top_p)
	items.append(item_frequency)
	items.append(item_presence)
	items.append(item_edit)
	items.append(item_system)
	items.append(item_add_sysmem)
	items.append(item_del_sysmem)
	items.append(item_tts)

	for item in items:
		for command in item.get('commands', []):
			handler_helps[command] = item.get('help', '')
			handlers[command] = item.get('fun', void_func)


populate()



from ..OwegaFun import existingFunctions
from ..config import baseConf
from ..utils import info_print
from ..OwegaSession import OwegaSession as ps


# enables/disables command execution
def handle_commands(temp_file, messages, given="", temp_is_temp=False):
	if given.lower() in ["on", "true", "enable", "enabled"]:
		baseConf["commands"] = True
		existingFunctions.enableGroup("utility.system")
		info_print("Command execution enabled.")
		return messages

	if given.lower() in ["off", "false", "disable", "disabled"]:
		baseConf["commands"] = False
		existingFunctions.disableGroup("utility.system")
		info_print("Command execution disabled.")
		return messages

	baseConf["commands"] = (not baseConf.get("commands", False))
	if baseConf.get("commands", False):
		existingFunctions.enableGroup("utility.system")
		info_print("Command execution enabled.")
	else:
		existingFunctions.disableGroup("utility.system")
		info_print("Command execution disabled.")
	return messages


item_commands = {
	"fun": handle_commands,
	"help": "toggles command execution / file creation",
	"commands": ["commands"],
}



OwegaLicense = """
#            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#                    Version 2, December 2004
#
# Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
#
# Everyone is permitted to copy and distribute verbatim or modified
# copies of this license document, and changing it is allowed as long
# as the name is changed.
#
#            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
#
#  0. You just DO WHAT THE FUCK YOU WANT TO.
"""



from ..config import baseConf
from ..utils import info_print
from ..OwegaSession import OwegaSession as ps


# enables/disables the TTS
def handle_tts(temp_file, messages, given="", temp_is_temp=False):
	if given.lower() in ["on", "true", "enable", "enabled"]:
		baseConf["tts_enabled"] = True
		info_print("Text-to-speech enabled.")
		return messages

	if given.lower() in ["off", "false", "disable", "disabled"]:
		baseConf["tts_enabled"] = False
		info_print("Text-to-speech disabled.")
		return messages

	baseConf["tts_enabled"] = (not baseConf.get("tts_enabled", False))
	if baseConf.get("tts_enabled", False):
		info_print("Text-to-speech enabled.")
	else:
		info_print("Text-to-speech disabled.")
	return messages


item_tts = {
	"fun": handle_tts,
	"help": "enables/disables the TTS",
	"commands": ["tts"],
}



from .config import baseConf
from .conversation import Conversation
from .OwegaFun import existingFunctions, connectLTS, functionlist_to_toollist
import time
import openai
import json5 as json
import re
from .utils import debug_print


# sometimes, GPT will give back invalid json.
# this function tries to make it valid
def convert_invalid_json(invalid_json):
	def replace_content(match):
		content = match.group(1)
		content = (
			content
			.replace('"', '\\"')
			.replace("\n", "\\n")
		)
		return f'"{content}"'
	valid_json = re.sub(r'`([^`]+)`', replace_content, invalid_json)
	return valid_json


# ask openAI a question
# TODO: comment a lot more
def ask(
	prompt: str = "",
	messages: Conversation = Conversation(),
	model=baseConf.get("model", ""),
	temperature=baseConf.get("temperature", 0.8),
	max_tokens=baseConf.get("max_tokens", 3000),
	function_call="auto",
	temp_api_key="",
	temp_organization="",
	top_p=baseConf.get("top_p", 1.0),
	frequency_penalty=baseConf.get("frequency_penalty", 0.0),
	presence_penalty=baseConf.get("presence_penalty", 0.0),
):
	if baseConf.get("debug", False):
		bc = baseConf.copy()
		bc["api_key"] = "REDACTED"
		debug_print(f"{bc}", True)
	connectLTS(messages.add_memory, messages.remove_memory, messages.edit_memory)
	old_api_key = openai.api_key
	old_organization = openai.organization
	if (prompt):
		messages.add_question(prompt)
	else:
		prompt = messages.last_question()
	if isinstance(function_call, bool):
		if function_call:
			function_call = "auto"
		else:
			function_call = "none"
	response = False
	while (not response):
		try:
			if (temp_api_key):
				openai.api_key = temp_api_key
			if (temp_organization):
				openai.organization = temp_organization
			response = openai.chat.completions.create(
				model=model,
				temperature=temperature,
				max_tokens=max_tokens,
				top_p=top_p,
				frequency_penalty=frequency_penalty,
				presence_penalty=presence_penalty,
				messages=messages.get_messages(),
				tools=functionlist_to_toollist(
					existingFunctions.getEnabled()),
				tool_choice=function_call,
			)
			if (temp_api_key):
				openai.api_key = old_api_key
			if (temp_organization):
				openai.organization = old_organization
		except openai.BadRequestError:
			messages.shorten()
		except openai.InternalServerError:
			print("[Owega] Service unavailable...", end="")
			time.sleep(1)
			print(" Retrying now...")
	# do something with the response
	message = response.choices[0].message
	while message.tool_calls is not None:
		try:
			for tool_call in message.tool_calls:
				tool_function = tool_call.function
				function_name = tool_function.name
				try:
					kwargs = json.loads(tool_function.arguments)
				except json.decoder.JSONDecodeError:
					unfixed = tool_function.arguments
					fixed = convert_invalid_json(unfixed)
					kwargs = json.loads(fixed)
				function_response = \
					existingFunctions.getFunction(function_name)(**kwargs)
				messages.add_function(function_name, function_response)
			response2 = False
			while not (response2):
				try:
					if (temp_api_key):
						openai.api_key = temp_api_key
					if (temp_organization):
						openai.organization = temp_organization
					response2 = openai.chat.completions.create(
						model=model,
						temperature=temperature,
						max_tokens=max_tokens,
						top_p=top_p,
						frequency_penalty=frequency_penalty,
						presence_penalty=presence_penalty,
						messages=messages.get_messages(),
						tools=functionlist_to_toollist(
							existingFunctions.getEnabled()),
						tool_choice=function_call,
					)
					if (temp_api_key):
						openai.api_key = old_api_key
					if (temp_organization):
						openai.organization = old_organization
				except openai.error.InvalidRequestError:
					messages.shorten()
				except openai.error.ServiceUnavailableError:
					print("[Owega] Service unavailable...", end="")
					time.sleep(1)
					print(" Retrying now...")
				message = response2.choices[0].message
		except Exception as e:
			print("Exception: " + str(e))
			print(message.tool_calls[0].function.name)
			print(message.tool_calls[0].function.arguments)
			break
	try:
		messages.add_answer(message.content.strip())
	except Exception as e:
		print("Exception: " + str(e))
	return messages



from ..utils import (
	do_quit,
	success_msg,
)
from ..OwegaSession import OwegaSession as ps


# quits the program, deleting temp_file
def handle_quit(temp_file, messages, given="", temp_is_temp=False):
	do_quit(
		success_msg(),
		temp_file=temp_file,
		is_temp=temp_is_temp,
		should_del=temp_is_temp
	)
	return messages


item_quit = {
	"fun": handle_quit,
	"help": "exits the program",
	"commands": ["quit", "exit"],
}



import prompt_toolkit as pt
from ..config import baseConf
from ..utils import (
	info_print,
	clrtxt,
)
from ..OwegaSession import OwegaSession as ps


# change temperature
def handle_temperature(temp_file, messages, given="", temp_is_temp=False):
	given = given.strip()
	try:
		new_temperature = float(given)
	except ValueError:
		info_print('Current temperature: '
			+ f'{baseConf.get("temperature", 1.0)}')
		info_print('New temperature value (0.0 - 2.0, defaults 0.8)')
		try:
			new_temperature = ps['float'].prompt(pt.ANSI(
				'\n' + clrtxt("magenta", " temperature ") + ': ')).strip()
		except (ValueError, KeyboardInterrupt, EOFError):
			info_print("Invalid temperature.")
			return messages
	baseConf["temperature"] = float(new_temperature)
	nv = baseConf.get('temperature', 0.0)
	if nv > 2.0:
		info_print('Temperature too high, capping to 2.0')
		baseConf["temperature"] = 2.0
	if nv < 0.0:
		info_print('Temperature negative, capping to 0.0')
		baseConf["temperature"] = 0.0
	info_print('Set temperature to '
		+ f'{baseConf.get("temperature", 0.0)}')
	return messages


item_temperature = {
	"fun": handle_temperature,
	"help": "sets the temperature (0.0 - 1.0, defaults 0.8)",
	"commands": ["temperature"],
}



#!/usr/bin/env python3
# Import the necessary modules
import openai
import os
import json5 as json
import getpass
import sys
import time
import re
import argparse
import prompt_toolkit as pt
import tempfile
from .changelog import OwegaChangelog
from .license import OwegaLicense
from .config import baseConf, get_conf, list_models
from .OwegaFun import existingFunctions, connectLTS, functionlist_to_toollist
from .OweHandlers import handlers, handle_help, handler_helps
from .utils import (
	get_home_dir,
	get_temp_file,
	info_print,
	do_quit,
	success_msg,
	clrtxt,
	print_help,
	estimated_tokens,
	play_opus,
)
from .conversation import Conversation, Conversation_from
from .ask import ask
from .OwegaSession import OwegaSession as ps


# get a copy of owega's config
def get_oc_conf():
	return baseConf.copy()


# main interaction loop
def user_interaction_loop(temp_file="", input_file="", temp_is_temp=False):
	if not temp_file:
		temp_is_temp = True
		temp_file = get_temp_file()

	# creates Conversation object and populate it
	messages = Conversation()
	connectLTS(
		messages.add_memory,
		messages.remove_memory,
		messages.edit_memory
	)
	if input_file:
		messages.load(input_file)

	# sets the input prompt
	input_prompt = '\n  ' + clrtxt("yellow", " USER ") + ": "

	# bootup info
	info_print("===== Owega =====")
	info_print(f"Owega v{OwegaChangelog.version}")
	info_print('Type "/help" for help')
	info_print(f"Default model is {baseConf.get('model', '')}")
	info_print(f"temp file is {temp_file}")

	# API key detection
	if baseConf.get("api_key", "").startswith("sk-"):
		openai.api_key = baseConf.get("api_key", "")
	else:
		# if key not in config: ask for key only if not already set (ie envvar)
		try:
			if not openai.api_key.startswith("sk-"):
				openai.api_key = getpass.getpass(prompt="OpenAI API Key: ")
		except AttributeError:
			openai.api_key = getpass.getpass(prompt="OpenAI API Key: ")
		baseConf["api_key"] = openai.api_key

	# Organization detection
	if baseConf.get("organization", "").startswith("org-"):
		openai.organization = baseConf.get("organization", "")

	# main interaction loop:
	while True:
		# save temp file
		messages.save(temp_file)

		# get user input, and strip it (no excess spaces / tabs / newlines
		user_input = ps['main'].prompt(pt.ANSI(input_prompt)).strip()

		command_found = False
		if user_input.startswith('/'):
			uinp_spl = user_input.split(' ')
			given = ' '.join(uinp_spl[1:])
			command = uinp_spl[0][1:]
			if command in handlers.keys():
				command_found = True
				current_handler = handlers.get(command, handle_help)
				messages = current_handler(
					temp_file,
					messages,
					given,
					temp_is_temp
				)
		if not command_found:
			if baseConf.get("estimation", False):
				etkn = estimated_tokens(
					user_input,
					messages,
					functionlist_to_toollist(existingFunctions.getEnabled())
				)
				cost_per_token = (
					0.03
					if 'gpt-4' in baseConf.get("model", "")
					else 0.003
				) / 1000
				cost = cost_per_token * etkn
				print(f"\033[37mestimated tokens: {etkn}\033[0m")
				print(f"\033[37mestimated cost: {cost:.5f}\033[0m")
			if baseConf.get("debug", False):
				pre_time = time.time()
			messages = ask(
				prompt=user_input,
				messages=messages,
				model=baseConf.get("model", ''),
				temperature=baseConf.get("temperature", 0.8),
				max_tokens=baseConf.get("max_tokens", 3000),
				top_p=baseConf.get("top_p", 1.0),
				frequency_penalty=baseConf.get("frequency_penalty", 0.0),
				presence_penalty=baseConf.get("presence_penalty", 0.0)
			)
			if baseConf.get("debug", False):
				post_time = time.time()
				print(f"\033[37mrequest took {post_time-pre_time:.3f}s\033[0m")

			# Print the generated response
			print()
			print(' ' + clrtxt("magenta", " Owega ") + ": ")
			print()
			print(messages.last_answer())

			if baseConf.get('tts_enabled', False):
				tmpfile = tempfile.NamedTemporaryFile(
					prefix="owegatts.",
					suffix=".opus",
					delete=False
				)
				tmpfile.close()
				tts_answer = openai.audio.speech.create(
					model='tts-1',
					voice='nova',
					input=messages.last_answer()
				)
				if baseConf.get("debug", False):
					posttts_time = time.time()
					print("\033[37mrequest took "
						+ f"{posttts_time-pre_time:.3f}s\033[0m")
				tts_answer.stream_to_file(tmpfile.name)
				play_opus(tmpfile.name)
				os.remove(tmpfile.name)


# parse command-line arguments
def parse_args():
	parser = argparse.ArgumentParser(description="Owega main application")
	parser.add_argument("-d", "--debug", action='store_true',
		help="Enable debug output")
	parser.add_argument("-c", "--changelog", action='store_true',
		help="Display changelog and exit")
	parser.add_argument("-l", "--license", action='store_true',
		help="Display license and exit")
	parser.add_argument("-v", "--version", action='store_true',
		help="Display version and exit")
	parser.add_argument("-f", "--config-file", type=str,
		help="Specify path to config file")

	parser.add_argument("-i", "--history", type=str,
		help="Specify the history file to import")

	parser.add_argument("-a", "--ask", type=str,
		help="Asks a question directly from the command line")

	parser.add_argument("-o", "--output", type=str,
		help="Saves the history to the specified file")

	parser.add_argument("-t", "--tts", action='store_true',
		help="Enables TTS generation when asking")
	parser.add_argument("-s", "--ttsfile", type=str,
		help="Outputs a generated TTS file single-ask mode")

	parser.add_argument("-T", "--training", action='store_true',
		help="outputs training data from -i file")
	parser.add_argument("-e", "--estimate", action='store_true',
		help="shows estimate token usage / cost from a request from -i file")

	return parser.parse_args()


# finds if any element in the list a is present in the list b
def is_la_in_lb(a, b):
	for e in a:
		if e in b:
			return True
	return False


# ask a single question (with a new context)
def single_ask(
	user_prompt,
	temp_file: str = "",
	input_file: str = "",
	temp_is_temp: bool = False,
	should_print: bool = False
):
	if not temp_file:
		temp_is_temp = True
		temp_file = get_temp_file()
	messages = Conversation()
	connectLTS(
		messages.add_memory,
		messages.remove_memory,
		messages.edit_memory
	)
	if input_file:
		messages.load(input_file)
	messages = ask(
		prompt=user_prompt,
		messages=messages,
		model=baseConf.get("model", ''),
		temperature=baseConf.get("temperature", 0.8),
		max_tokens=baseConf.get("max_tokens", 3000),
		top_p=baseConf.get('top_p', 1.0),
		frequency_penalty=baseConf.get('frequency_penalty', 0.0),
		presence_penalty=baseConf.get('presence_penalty', 0.0)
	)
	if should_print:
		print(messages.last_answer())
	if baseConf.get('tts_enabled', False):
		tmpfile = tempfile.NamedTemporaryFile(
			prefix="owegatts.",
			suffix=".opus",
			delete=False
		)
		tmpfile.close()
		tts_answer = openai.audio.speech.create(
			model='tts-1',
			voice='nova',
			input=messages.last_answer()
		)
		tts_answer.stream_to_file(tmpfile.name)
		play_opus(tmpfile.name)
		os.remove(tmpfile.name)
	return messages.last_answer()
	if not temp_is_temp:
		messages.save(temp_file)


# main function, runs if __name__ is "__main__" ;)
def main():
	args = parse_args()

	if (args.debug):  # bypass before loading conf
		baseConf["debug"] = True

	if args.changelog:
		print(OwegaChangelog.log)
	if args.license:
		print(OwegaLicense)
	if args.version:
		print(f"Owega v{OwegaChangelog.version}")
	if (args.changelog or args.license or args.version):
		do_quit(value=1)
	if (args.training and not args.history):
		do_quit("Can't generate training data without a history", value=1)
	if args.training:
		msgs = Conversation_from(args.history)
		print(msgs.generate_training())
		sys.exit(0)
	if (args.estimate and not args.history):
		do_quit(
			"Can't estimate token consumption/cost without a history", value=1)
	if args.estimate:
		msgs = Conversation_from(args.history)
		etkn = estimated_tokens(
			'',
			msgs,
			''
		)
		cost_per_token = (
			0.003
		) / 1000
		cost = cost_per_token * etkn
		print(f"estimated tokens: {etkn}")
		print(f"estimated cost: {cost:.5f}$ (gpt-3) / {cost*10:.5f}$ (gpt-4)")
		sys.exit(0)

	input_history = ""
	if (args.history):
		input_history = args.history

	temp_file = get_temp_file()
	temp_is_temp = True
	if (args.output):
		temp_is_temp = False
		temp_file = args.output

	get_conf(args.config_file)
	if baseConf.get("commands", False):
		existingFunctions.enableGroup("utility.system")
	else:
		existingFunctions.disableGroup("utility.system")

	if (args.debug):  # bypass after loading conf
		baseConf["debug"] = True

	if (args.tts):
		baseConf["tts_enabled"] = True

	if (args.ask):
		answer = single_ask(
			args.ask,
			temp_file,
			input_history,
			temp_is_temp,
			True
		)
		if (args.ttsfile):
			tts_answer = openai.audio.speech.create(
				model="tts-1",
				voice="nova",
				input=answer
			)
			if (("opus" not in args.ttsfile)
				and ("mp3" not in args.ttsfile)
				and ("aac" not in args.ttsfile)
				and ("flac" not in args.ttsfile)):
				args.ttsfile = args.ttsfile + '.opus'
			tts_answer.stream_to_file(args.ttsfile)
	else:
		try:
			user_interaction_loop(
				temp_file=temp_file,
				input_file=input_history,
				temp_is_temp=temp_is_temp
			)
		except EOFError:
			do_quit(
				success_msg(),
				temp_file=temp_file,
				is_temp=temp_is_temp,
				should_del=temp_is_temp
			)
		except KeyboardInterrupt:
			do_quit(
				success_msg(),
				temp_file=temp_file,
				is_temp=temp_is_temp,
				should_del=False
			)


if __name__ == "__main__":
	main()



from ..utils import genconfig
from ..OwegaSession import OwegaSession as ps


# generates config file
def handle_genconf(temp_file, messages, given="", temp_is_temp=False):
	genconfig()
	return messages


item_genconf = {
	"fun": handle_genconf,
	"help": "generates a sample config file",
	"commands": ["genconf"],
}
